from ..backend.NRV_Class import NRV_class
from ..backend.log_interface import rise_error



class eit_inverse(NRV_class):
    """
    Class allowing image reconstruction from ``eit_forward`` results. Based on the `pyEIT` algorithms.
    """
    def __init__():
        _protocol = None
        rise_error(NotImplementedError)



